import ij.*;
import ij.process.*;
import ij.gui.*;
import ij.util.*;
import java.awt.*;
import java.awt.event.*;
import ij.plugin.*;
import ij.plugin.frame.*;
import ij.measure.*;
import java.util.*;
import java.util.Arrays;
import java.util.concurrent.atomic.AtomicInteger;

    /**
      This plugin generates Frame-Dimension projections of a Hyperstack.

      @author Fred Damen <fred@damen.org>

      Version History:
      2018-04-01: Created
      2019-04-01: Updated
      2020-04-01: Updated
    */

public class F_Project implements PlugIn {

   /** gui
    */
   public void run(String arg) {
      ImagePlus imp = IJ.getImage();
      run(imp,arg);
      }
   /** gui
    */
   public void run(ImagePlus imp, String arg) {
      if (IJ.versionLessThan("1.52b")) 
         return;

      final GenericDialog gd = new GenericDialog("Frame Projector");
      Vector<String> itv = new Vector<String>();
      int defind = -1;
      final int ids[] = WindowManager.getIDList();
      for(int id : ids) {
         if (id == imp.getID()) defind = itv.size();
         itv.add(WindowManager.getImage(id).getTitle());
         }
      gd.addChoice("Multi Frame Image:", itv.toArray(new String[0]), "");
      String str = "";
      if (imp.getType() != ImagePlus.GRAY32) str += "     Not 32-bit float... \n";
      if (imp.getNFrames() < 2)              str += "     Not a HyperStack...\n";
      gd.addMessage(str.equals("") ? " \n " : "Caveat emptor: \n"+str);
      final MultiLineLabel msgw = (MultiLineLabel)gd.getMessage();
    
      ((Choice)gd.getChoices().get(0)).select(defind);

      String slices = "1-"+imp.getNSlices()+"-1";
      gd.addStringField("Slices: ", slices); gd.addToSameRow(); gd.addMessage("from-to-by\n'all'");
      final TextField stfw = (TextField)gd.getStringFields().get(0);
      stfw.setColumns(25);
      String frames = "1-"+imp.getNFrames()+"-1";
      gd.addStringField("Frames: ", frames); gd.addToSameRow(); gd.addMessage("from-to-by\n'all' or 'label var' or 'repeats var'");
      final TextField ftfw = (TextField)gd.getStringFields().get(1);
      ftfw.setColumns(25);
  
      ((Choice)gd.getChoices().get(0)).addItemListener(new ItemListener() {
         @Override public void itemStateChanged(ItemEvent e) {
            ImagePlus imp = WindowManager.getImage(ids[((Choice)e.getSource()).getSelectedIndex()]);
            String str = "";
            if (imp.getType() != ImagePlus.GRAY32) str += "     Not 32-bit float... \n";
            if (imp.getNFrames() < 2)              str += "     Not a HyperStack...\n";
            msgw.setText(str.equals("") ? "\n \n \n " : "Caveat emptor: \n"+str);
            stfw.setText("1-"+imp.getNSlices()+"-1");
            ftfw.setText("1-"+imp.getNFrames()+"-1");
            gd.pack();
            };
         });

      gd.addCheckbox("Transpose ?", false); gd.addToSameRow(); gd.addMessage("Z_Projector");

      gd.addMessage("Projections...");
      gd.addCheckbox("Sum",       false);
      gd.addCheckbox("Mean",      false);
      gd.addCheckbox("StdDev",    false);
      gd.addCheckbox("ZScore",    false); gd.addToSameRow(); gd.addMessage("(aka SNR)");
      gd.addCheckbox("CV",        false); gd.addToSameRow(); gd.addMessage("Coefficient of Variation");
      gd.addCheckbox("Minimum",   false);
      gd.addCheckbox("Maximum",   false);
      gd.addCheckbox("Median",    false);
      gd.addCheckbox("Center",    false); gd.addToSameRow(); gd.addMessage("Resilient to outliers");
      gd.addCheckbox("Product",   false); gd.addToSameRow(); gd.addMessage("Product (pre nth-rooted)");
      gd.addCheckbox("GeoMean",   false); gd.addToSameRow(); gd.addMessage("Geometric Mean (pre nth-rooted)");
      gd.addCheckbox("SumSq",     false); gd.addToSameRow(); gd.addMessage("Sum of Squared");
      gd.addCheckbox("Magnitude", false); gd.addToSameRow(); gd.addMessage("Pythagoras");
      gd.addCheckbox("LinReg",    false); gd.addToSameRow(); gd.addMessage("Linear Least Squares Regression");
      gd.addCheckbox("TheilSen",  false); gd.addToSameRow(); gd.addMessage("Theil Sen Linear Regression");
      gd.addCheckbox("TheilSen2", false); gd.addToSameRow(); gd.addMessage("Theil Sen Quadratic Regression");
      gd.addCheckbox("Centroid",  false);

      String[] fitfunChoices = new String[CurveFitter.fitList.length+1];
      fitfunChoices[0] = "none";
      for(int i=1; i<fitfunChoices.length; i++)
         fitfunChoices[i] = CurveFitter.fitList[CurveFitter.sortedTypes[i-1]];
      gd.addChoice("Fit", fitfunChoices, fitfunChoices[0]);

      gd.addCheckbox("Deviation", false); gd.addToSameRow(); gd.addMessage("from aforechecked");
      
      gd.showDialog();
      if (gd.wasCanceled()) return;

      imp = WindowManager.getImage(ids[gd.getNextChoiceIndex()]);

      slices = gd.getNextString();
      frames = gd.getNextString();
      boolean trans = gd.getNextBoolean();
      boolean[] flags = new boolean[]{ false,
                                       gd.getNextBoolean(),
                                       gd.getNextBoolean(),
                                       gd.getNextBoolean(),
                                       gd.getNextBoolean(),
                                       gd.getNextBoolean(),
                                       gd.getNextBoolean(),
                                       gd.getNextBoolean(),
                                       gd.getNextBoolean(),
                                       gd.getNextBoolean(),
                                       gd.getNextBoolean(),
                                       gd.getNextBoolean(),
                                       gd.getNextBoolean(),
                                       gd.getNextBoolean(),
                                       gd.getNextBoolean(),
                                       gd.getNextBoolean(),
                                       gd.getNextBoolean(),
                                       gd.getNextBoolean() };
      flags[0] = gd.getNextBoolean();

      int fitfun = gd.getNextChoiceIndex() - 1;

      try { 
         for(ImagePlus result : compute(imp, slices, frames, flags, fitfun, null, null, trans))
            if (result != null) result.show();
         }
      catch(IllegalArgumentException e) { 
         IJ.error(e.getMessage()); 
         return; 
         }
      }

   /** Swap z direction for f direction.
    */
   public static ImagePlus transpose(ImagePlus imp) {
      return HyperStackConverter.toHyperStack(imp, 1, imp.getNFrames(), imp.getNSlices(), "xyctz", "Grayscale");
      }
   /** Subset then swap z direction for f direction.
    */
   public static ImagePlus transpose(ImagePlus imp, String slices, String frames) {
      ImagePlus tmpimp = subhyperstack(imp,"1",slices,frames);
      return HyperStackConverter.toHyperStack(tmpimp, 1, tmpimp.getNFrames(), tmpimp.getNSlices(), "xyctz", "Grayscale");
      }
   /** Swap z direction for f direction.
    */
   public static ImagePlus[] transpose(ImagePlus imps[]) {
      for(ImagePlus imp : imps)
         if (imp != null)
            imp = HyperStackConverter.toHyperStack(imp, 1, imp.getNFrames(), imp.getNSlices(), "xyctz", "Grayscale");
      return imps;
      }

   /** Extract variable/value pairs from frame's slice label.
    */
   public static TreeMap<String,Double> getVars(ImagePlus imp, int f) {
      TreeMap<String,Double> cvv = new TreeMap<String,Double>();
      try {
         for(String vs : imp.getStack().getSliceLabel(imp.getStackIndex(1,1,f)).split("\n",2)[0].split(",|;")) {
             String[] vv = vs.split(" = |=| ",2);
             cvv.put(vv[0], new Double(vv[1]));
             }
         }
      catch(Throwable t) { 
         return null;
         }
      return cvv;
      }
   /** Extract variable/values from all frames' slice labels.
    *  see Frame_Slider
    */
   public static TreeMap<String,TreeMap<Double,Double>> getVars(ImagePlus imp) {
      TreeMap<String,TreeMap<Double,Double>> vMap = new TreeMap<String,TreeMap<Double,Double>>();
      ImageStack is = imp.getStack();
      int nf = imp.getNFrames();
      try {
         for(int f=0; f<nf; f++) {
            for(String vs : is.getSliceLabel(imp.getStackIndex(1,1,f+1)).split("\n",2)[0].split(",|;")) {
               String[] vv = vs.split(" = |=| ",2);
               if (!vMap.containsKey(vv[0]))
                  vMap.put(vv[0], new TreeMap<Double,Double>());
               if (!vMap.get(vv[0]).containsKey(Double.valueOf(vv[1])))
                  vMap.get(vv[0]).put(Double.valueOf(vv[1]),(double)vMap.get(vv[0]).size()+1);
               }
            }
         }
      catch(Throwable e) {
         return vMap;
         }
      return vMap;
      }
   /** Find the frame (1 based) that contains the variable/value pairs in its slice label. Subsets are allowed.
    * see Frame_Slider
    */
   public static int findFrame(ImagePlus imp, TreeMap<String,Double> varval) {
      ImageStack is = imp.getStack();
      int nf = imp.getNFrames();
      try {
         for(int f=0; f<nf; f++) {
            boolean found = true;
            for(String vs : is.getSliceLabel(imp.getStackIndex(1,1,f+1)).split("\n",2)[0].split(",|;")) {
               String[] vv = vs.split(" = |=| ",2);
               if (varval.get(vv[0]) != Double.valueOf(vv[1]).doubleValue()) {
                  found = false;
                  break;
                  }
               }
            if (found) return f+1;
            }
         }
      catch(Throwable e) {
         return -1;
         }
      return -1;
      }

   /** Determine the subsets (keys) that exist and the associated comma separated frames (values), where var is the repeater.
    *  see 'repeats var'
    */
   public static LinkedHashMap<String,String> overRepeats(ImagePlus imp, String var) {
      LinkedHashMap<String,String> map = new LinkedHashMap<String,String>();
      try {
         ImageStack is = imp.getStack();
         int nf = imp.getNFrames();
         for(int f=0; f<nf; f++) { 
            String key="";
            for(String vs : is.getShortSliceLabel(imp.getStackIndex(1,1,f+1)).split(",|;"))
               if (!vs.split(" = |=| ",2)[0].equals(var)) key += vs+";";
            map.put(key,map.getOrDefault(key,"")+","+(f+1));
            }
         for(Map.Entry<String,String> e : map.entrySet())
            e.setValue(e.getValue().substring(1));
         }
      catch(Throwable e) {
         return null;
         }
      return map;
      }
   /** Remove superfluous variable var from all clice labels
    */
   public static void removeVarFromLabels(ImagePlus imp, String var) {
      int N = imp.getNFrames()*imp.getNSlices();
      ImageStack is = imp.getStack();
      for(int s=0; s<N; s++) {
         String[] sl = is.getSliceLabel(s+1).split("\n",2);
         String nsl = "";
         for(String vs : sl[0].split(",|;"))
            if (!vs.split(" = |=| ",2)[0].equals(var))
               nsl += vs+";";
         is.setSliceLabel(nsl.replaceAll(";+$","")+"\n"+sl[1],s+1);
         }
      }
   /** Replace all the values for var with val in all the slice labels.
    */
   public static void replaceVarValWithInLabel(ImagePlus imp, String var, String val) {
      int NS = imp.getImageStackSize();
      ImageStack is = imp.getStack();
      for(int s=0; s<NS; s++) {
         String[] sl = is.getSliceLabel(s+1).split("\n",2);
         String nsl = "";
         for(String vs : sl[0].split(",|;")) {
            String[] vv = vs.split(" = |=| ",2);
            nsl += (vv[0].equals(var) ? vv[0]+" "+val : vs)+";";
            }
         is.setSliceLabel(nsl.replaceAll(";+$","")+"\n"+sl[1],s+1);
         }
      }
   /** Determines position of var in the slice label variables. -1 not found or error.
    */
   public static int findVarInLabel(ImagePlus imp, String var) {
      try {
         ImageStack is = imp.getStack();
         int iv = 0;
         for(String str : is.getShortSliceLabel(imp.getStackIndex(1,1,1)).split(",|;")) {
            if (str.split(" = |=| ",2)[0].equals(var))
               return iv;
            iv++;
            }
         }
      catch(Throwable e) {
         return -1;
         }
      return -1;
      }

   /** Do not use.
    */
   public static TreeMap<Double,String> sortFramesByLabel(ImagePlus imp) {
      return sortFramesByLabel(imp,0);
      }
   /** Determine frames (values - comma separated) that corrosponds to each var values (keys).
    */
   public static TreeMap<Double,String> sortFramesByLabel(ImagePlus imp, String var) {
      return sortFramesByLabel(imp,findVarInLabel(imp,var));
      }
   private static TreeMap<Double,String> sortFramesByLabel(ImagePlus imp, int iv) {
      TreeMap<Double,String> map = new TreeMap<Double,String>();
      try {
         ImageStack is = imp.getStack();
         int nf = imp.getNFrames();
         for(int f=0; f<nf; f++) {
            String str = is.getSliceLabel(imp.getStackIndex(1,1,f+1)).split("\n",2)[0].split(",|;")[iv].split(" = |=| ",2)[1];
            Double d = Double.valueOf(str).doubleValue();
            map.put(d,map.getOrDefault(d,"")+","+(f+1));
            }
         for(Map.Entry<Double,String> e : map.entrySet())
            e.setValue(e.getValue().substring(1));
         }
      catch(Throwable e) {
         return null;
         }
      return map;
      }

   /** Extract all frames that have the variable / value pair var/val.
    */
   public static ImagePlus extractFramesByVar(ImagePlus imp, String var, double val) {
      TreeMap<Double,String> fmap = sortFramesByLabel(imp, var);
      return subhyperstack(imp,"1-1","1-"+imp.getNSlices(),fmap.get(val));
      }
   /** Extract the frames (as an ImagePlus) that corresponds to each value (key) for slice label variable var.
    */
   public static TreeMap<Double,ImagePlus> extractFramesByVar(ImagePlus imp, String var) {
      TreeMap<Double,ImagePlus> imap = new TreeMap<Double,ImagePlus>();
      TreeMap<Double,String> fmap = sortFramesByLabel(imp, var);
      for(Map.Entry<Double,String> e : fmap.entrySet())
         imap.put(e.getKey(), subhyperstack(imp,"1-1","1-"+imp.getNSlices(),e.getValue()));
      return imap;
      }
   /** Rearrange frame order to that of increasing var values.
    */
   public static ImagePlus sortFramesByVar(ImagePlus imp, String var) {
      TreeMap<Double,String> map = sortFramesByLabel(imp,var);
      String str = "";
      for(Map.Entry<Double,String> e : map.entrySet())
         str += e.getValue()+",";
      return subhyperstack(imp,"1-1","1-"+imp.getNSlices(),str.substring(1));
      }

   /** Same as extractFramesByVar ???
    */
   public static TreeMap<Double,ImagePlus> byLabel(ImagePlus imp, String var) {
      TreeMap<Double,ImagePlus> mapi = new TreeMap<Double,ImagePlus>();
      try {
         ImageStack is = imp.getStack();
         int iv = findVarInLabel(imp,var);
         if (iv==-1) {
            mapi.put(Double.NaN,imp);
            return mapi;
            }
         int nf = imp.getNFrames();
         TreeMap<Double,String> maps = new TreeMap<Double,String>();
         for(int f=0; f<nf; f++) {
            String str = is.getSliceLabel(imp.getStackIndex(1,1,f+1)).split("\n",2)[0].split(",|;")[iv].split(" = |=| ",2)[1];
            Double d = Double.valueOf(str).doubleValue();
            maps.put(d,maps.getOrDefault(d,"")+","+(f+1));
            }
         for(Map.Entry<Double,String> e : maps.entrySet())
            mapi.put(e.getKey(),subhyperstack(imp,"1-1","1-"+imp.getNSlices()+"-1",e.getValue().substring(1)));
         }
      catch(Throwable e) {
         return null;
         }
      return mapi;
      }

   public static ImagePlus sum(ImagePlus imp)       { return sum(      imp, "all", "all"); }
   public static ImagePlus mean(ImagePlus imp)      { return mean(     imp, "all", "all"); }
   public static ImagePlus stddev(ImagePlus imp)    { return stddev(   imp, "all", "all"); }
   public static ImagePlus zscore(ImagePlus imp)    { return zscore(   imp, "all", "all"); }
   public static ImagePlus cv(ImagePlus imp)        { return cv(       imp, "all", "all"); }
   public static ImagePlus min(ImagePlus imp)       { return min(      imp, "all", "all"); }
   public static ImagePlus max(ImagePlus imp)       { return max(      imp, "all", "all"); }
   public static ImagePlus median(ImagePlus imp)    { return median(   imp, "all", "all"); }
   public static ImagePlus center(ImagePlus imp)    { return center(   imp, "all", "all"); }
   public static ImagePlus product(ImagePlus imp)   { return product(  imp, "all", "all"); }
   public static ImagePlus geomean(ImagePlus imp)   { return geomean(  imp, "all", "all"); }
   public static ImagePlus sumsq(ImagePlus imp)     { return sumsq(    imp, "all", "all"); }
   public static ImagePlus magnitude(ImagePlus imp) { return magnitude(imp, "all", "all"); }
   public static ImagePlus linreg(ImagePlus imp)    { return linreg(   imp, "all", "all"); }
   public static ImagePlus theilsen(ImagePlus imp)  { return theilsen( imp, "all", "all"); }
   public static ImagePlus theilsen2(ImagePlus imp) { return theilsen2(imp, "all", "all"); }
   public static ImagePlus centroid(ImagePlus imp)  { return centroid( imp, "all", "all"); }

   public static ImagePlus fit(ImagePlus imp, int fitfun)       { return fit(imp, "all", "all", fitfun); }
   //public static ImagePlus fit(ImagePlus imp, String fitfunstr) { return fit(imp, "all", "all", fitfunstr); }

   public static ImagePlus sum(ImagePlus imp, String slices, String frames) {
      return compute(imp, slices, frames, new boolean[]{false,true})[0];
      }
   public static ImagePlus mean(ImagePlus imp, String slices, String frames) {
      return compute(imp, slices, frames, new boolean[]{false,false,true})[1];
      }
   public static ImagePlus stddev(ImagePlus imp, String slices, String frames) {
      return compute(imp, slices, frames, new boolean[]{false,false,false,true})[2];
      }
   public static ImagePlus zscore(ImagePlus imp, String slices, String frames) {
      return compute(imp, slices, frames, new boolean[]{false,false,false,false,true})[3];
      }
   public static ImagePlus cv(ImagePlus imp, String slices, String frames) {
      return compute(imp, slices, frames, new boolean[]{false,false,false,false,false,true})[4];
      }
   public static ImagePlus min(ImagePlus imp, String slices, String frames) {
      return compute(imp, slices, frames, new boolean[]{false,false,false,false,false,false,true})[5];
      }
   public static ImagePlus max(ImagePlus imp, String slices, String frames) {
      return compute(imp, slices, frames, new boolean[]{false,false,false,false,false,false,false,true})[6];
      }
   public static ImagePlus median(ImagePlus imp, String slices, String frames) {
      return compute(imp, slices, frames, new boolean[]{false,false,false,false,false,false,false,false,true})[7];
      }
   public static ImagePlus center(ImagePlus imp, String slices, String frames) {
      return compute(imp, slices, frames, new boolean[]{false,false,false,false,false,false,false,false,false,true})[8];
      }
   public static ImagePlus product(ImagePlus imp, String slices, String frames) {
      return compute(imp, slices, frames, new boolean[]{false,false,false,false,false,false,false,false,false,false,true})[9];
      }
   public static ImagePlus geomean(ImagePlus imp, String slices, String frames) {
      return compute(imp, slices, frames, new boolean[]{false,false,false,false,false,false,false,false,false,false,false,true})[10];
      }
   public static ImagePlus sumsq(ImagePlus imp, String slices, String frames) {
      return compute(imp, slices, frames, new boolean[]{false,false,false,false,false,false,false,false,false,false,false,false,true})[11];
      }
   public static ImagePlus magnitude(ImagePlus imp, String slices, String frames) {
      return compute(imp, slices, frames, new boolean[]{false,false,false,false,false,false,false,false,false,false,false,false,false,true})[12];
      }
   public static ImagePlus linreg(ImagePlus imp, String slices, String frames) {
      return compute(imp, slices, frames, new boolean[]{false,false,false,false,false,false,false,false,false,false,false,false,false,false,true})[13];
      }
   public static ImagePlus theilsen(ImagePlus imp, String slices, String frames) {
      return compute(imp, slices, frames, new boolean[]{false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,true})[14];
      }
   public static ImagePlus theilsen2(ImagePlus imp, String slices, String frames) {
      return compute(imp, slices, frames, new boolean[]{false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,true})[15];
      }
   public static ImagePlus centroid(ImagePlus imp, String slices, String frames) {
      return compute(imp, slices, frames, new boolean[]{false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,false,true})[16];
      }

   /** For computes with deviations, a two element array instead of double length hyperstack.
    */
   public static ImagePlus[] splitDev(ImagePlus imp) {
      int NF = imp.getNFrames();
      int NZ = imp.getNSlices();
      return new ImagePlus[]{subhyperstack(imp,"1","1-"+NZ,"1-"+NF+"-2"),subhyperstack(imp,"1","1-"+NZ,"2-"+NF+"-2")};
      }
   public static ImagePlus[] meanDev(ImagePlus imp)    { return meanDev(  imp, "all", "all"); }
   public static ImagePlus[] medianDev(ImagePlus imp)  { return medianDev(imp, "all", "all"); }
   public static ImagePlus[] centerDev(ImagePlus imp)  { return centerDev(imp, "all", "all"); }
   public static ImagePlus[] geomeanDev(ImagePlus imp) { return geomeanDev(imp, "all", "all"); }

   public static ImagePlus[] meanDev(ImagePlus imp, String slices, String frames) {
      return splitDev(compute(imp, slices, frames, new boolean[]{true,false,true})[1]);
      }
   public static ImagePlus[] medianDev(ImagePlus imp, String slices, String frames) {
      return splitDev(compute(imp, slices, frames, new boolean[]{true,false,false,false,false,false,false,false,true})[7]);
      }
   public static ImagePlus[] centerDev(ImagePlus imp, String slices, String frames) {
      return splitDev(compute(imp, slices, frames, new boolean[]{true,false,false,false,false,false,false,false,false,true})[8]);
      }
   public static ImagePlus[] geomeanDev(ImagePlus imp, String slices, String frames) {
      return splitDev(compute(imp, slices, frames, new boolean[]{true,false,false,false,false,false,false,false,false,false,true})[10]);
      }

   public static ImagePlus fit(ImagePlus imp, String slices, String frames, int fitfun, ImagePlus mask) {
      return lastlast(compute(imp, slices, frames, new boolean[0], fitfun, null, mask));
      }
/*
   public static ImagePlus fit(ImagePlus imp, String slices, String frames, String fitfunstr, ImagePlus mask) throws IllegalArgumentException {
      for(int fitfun=0; fitfun<CurveFitter.fitList.length; fitfun++)
         if (CurveFitter.fitList[fitfun].equals(fitfunstr))
            return lastlast(compute(imp, slices, frames, new boolean[0], fitfun, null, mask));
      throw new IllegalArgumentException("Fitting Function Not found.");
      }
*/
   public static ImagePlus fit(ImagePlus imp, String slices, String frames, int fitfun) {
      return lastlast(compute(imp, slices, frames, new boolean[0], fitfun, null, null));
      }
/*
   public static ImagePlus fit(ImagePlus imp, String slices, String frames, String fitfunstr) throws IllegalArgumentException {
      return fit(imp, slices, frames, fitfunstr, (ImagePlus)null);
      }
*/
   private static ImagePlus lastlast(ImagePlus[] imps) {
      return imps[imps.length-1-1];
      }

   /** Interface for passing a user method to compute the projection.
    */
   public interface Compute {
      /** if x == null return new double[N]; // where N is the number of returned parameters
       *  else ...
       */
      public double[] compute(double[] x, double[] y);
      }
   /** returns hyperstack where each frame is one of he parameters.
    */
   public static ImagePlus compute(ImagePlus imp, Compute fp) { 
      return last(compute(imp, "all", "all", new boolean[0], -1, fp, null));
      }
   public static ImagePlus compute(ImagePlus imp, String slices, String frames, Compute fp) { 
      return last(compute(imp, slices, frames, new boolean[0], -1, fp, null));
      }
   /** compute method is not called on mask'ed out voxels.
    */
   public static ImagePlus compute(ImagePlus imp, Compute fp, ImagePlus mask) { 
      return last(compute(imp, "all", "all", new boolean[0], -1, fp, mask));
      }
   public static ImagePlus compute(ImagePlus imp, String slices, String frames, Compute fp, ImagePlus mask) { 
      return last(compute(imp, slices, frames, new boolean[0], -1, fp, mask));
      }

   private static ImagePlus last(ImagePlus[] imps) {
      return imps[imps.length-1];
      }

   /** n-tuple results returned as n-element array instead of n-frame hypderstack.
    */
   public static ImagePlus[] split(ImagePlus imp) {
      ImagePlus[] imps = new ImagePlus[imp.getNFrames()];
      for(int i=0; i<imps.length; i++)
         imps[i] = subhyperstack(imp,"1","1-"+imp.getNSlices(),""+(i+1));
      return imps;
      }

   /** parameter imps is purely syntatic sugar.
    */
   public static ImagePlus[] fit(ImagePlus imp, int fitfun, ImagePlus[] imps) {
      return split(fit(imp, fitfun));
      }
   public static ImagePlus[] fit(ImagePlus imp, String slices, String frames, int fitfun, ImagePlus[] imps) {
      return split(fit(imp,slices,frames,fitfun));
      }
/*
   public static ImagePlus[] fit(ImagePlus imp, String slices, String frames, String fitfunstr, ImagePlus[] imps) throws IllegalArgumentException {
      return split(fit(imp, slices, frames, fitfunstr));
      }
*/


   public static ImagePlus[] compute(ImagePlus imp, String slices, String frames, boolean[] flags) {
      return compute(imp, slices, frames, flags, -1, null);
      }

   /** parameter imps is purely syntatic sugar.
    */
   public static ImagePlus[] compute(ImagePlus imp, Compute fp, ImagePlus[] imps) {
      return split(compute(imp, fp));
      }
   public static ImagePlus[] compute(ImagePlus imp, String slices, String frames, Compute fp, ImagePlus[] imps) { 
      return split(compute(imp, slices, frames, fp));
      }
   public static ImagePlus[] compute(ImagePlus imp, Compute fp, ImagePlus mask, ImagePlus[] imps) {
      return split(compute(imp, fp, mask));
      }
   public static ImagePlus[] compute(ImagePlus imp, String slices, String frames, Compute fp, ImagePlus mask, ImagePlus[] imps) { 
      return split(compute(imp, slices, frames, fp, mask));
      }

   public static ImagePlus[] compute(ImagePlus imp, String slices, String frames, boolean[] flags, int fitfun) {
      return compute(imp, slices, frames, flags, fitfun, null);
      }

   /** transpose, project, transpose
    */
   public static ImagePlus[] compute(ImagePlus imp, String slices, String frames, boolean[] flags, int fitfun, Compute fp, ImagePlus mask, boolean trans) {
      if (trans) return transpose(compute(transpose(imp, slices, frames),"all","all",flags,fitfun,fp,null));
            else return compute(imp,slices,frames,flags,fitfun,fp,null);
      }

   /** mask not required.
    */
   public static ImagePlus[] compute(ImagePlus imp, String slices, String frames, boolean[] flags, int fitfun, Compute fp) {
      return compute(imp, slices, frames, flags, fitfun, fp, null);
      }

   /** The work horse.
    *  Should not call directly.
    */
   private static ImagePlus[] compute(ImagePlus imp, String slices, String frames, boolean[] flags, final int fitfun, final Compute fp, final ImagePlus mask) {

      if (imp == null) return null;

      if (slices.equals("all")) slices = "1-"+imp.getNSlices()+"-1";
      if (frames.equals("all")) frames = "1-"+imp.getNFrames()+"-1";
      if (frames.equals("label")) {
         TreeMap<Double,String> flmap = sortFramesByLabel(imp);
         if (flmap == null) return null;
         ImagePlus[] resimp = compute(imp,slices,flmap.firstEntry().getValue(),flags,fitfun,fp,mask);
         int i = 0;
         for(Map.Entry<Double,String> e : flmap.entrySet()) {
            if (i++==0) continue;
            int j=0;
            for(ImagePlus tmpimp : compute(imp,slices,e.getValue(),flags,fitfun,fp,mask)) {
               if (resimp[j] != null) {
                  resimp[j] = (new Concatenator()).concatenate(resimp[j],tmpimp,true);
                  resimp[j].setTitle(tmpimp.getTitle());
                  }
               j++;
               }
            }
         return resimp;
         }
      if (frames.startsWith("repeats")) {
         String var = frames.split(" ",2)[1];
         LinkedHashMap<String,String> frmap = overRepeats(imp,var);
         if (frmap == null) return null;
         ImagePlus resimp[] = null;
         for(Map.Entry<String,String> e : frmap.entrySet()) {
            if (resimp == null) {
               resimp = compute(imp,slices,e.getValue(),flags,fitfun,fp,mask);
               continue;
               }
            int j = 0;
            for(ImagePlus tmpimp : compute(imp,slices,e.getValue(),flags,fitfun,fp,mask)) {
               if (resimp[j] != null) {
                  resimp[j] = (new Concatenator()).concatenate(resimp[j],tmpimp,true);
                  resimp[j].setTitle(tmpimp.getTitle());
                  }
               j++;
               }
            }
         for(int i=0; i<resimp.length; i++)
            if (resimp[i] != null)
               removeVarFromLabels(resimp[i],var);
         return resimp;
         }


      final boolean doDev = (flags.length >  0) && flags[ 0];

      final boolean doSum = (flags.length >  1) && flags[ 1];
      final boolean doAve = (flags.length >  2) && flags[ 2];
      final boolean doSD  = (flags.length >  3) && flags[ 3];
      final boolean doZS  = (flags.length >  4) && flags[ 4];
      final boolean doCV  = (flags.length >  5) && flags[ 5];
      final boolean doMin = (flags.length >  6) && flags[ 6];
      final boolean doMax = (flags.length >  7) && flags[ 7];
      final boolean doMed = (flags.length >  8) && flags[ 8];
      final boolean doCen = (flags.length >  9) && flags[ 9];
      final boolean doPro = (flags.length > 10) && flags[10];
      final boolean doGeo = (flags.length > 11) && flags[11];
      final boolean doSSq = (flags.length > 12) && flags[12];
      final boolean doMag = (flags.length > 13) && flags[13];
      final boolean doLR  = (flags.length > 14) && flags[14];
      final boolean doTS  = (flags.length > 15) && flags[15];
      final boolean doTS2 = (flags.length > 16) && flags[16];
      final boolean doCEN = (flags.length > 17) && flags[17];

      final boolean doFit = fitfun >= 0;

      final boolean doCom = fp != null;

      ImageCalculator ic = new ImageCalculator();

      final ImagePlus subimp = subhyperstack(imp,"1-1",slices,frames);
      int NZ = subimp.getNSlices();
      int NF = subimp.getNFrames();

      ImagePlus tmpimp = subhyperstack(subimp,"1","1-"+NZ,"1");
      ImagePlus sumimp = (doSum|doAve|doSD|doZS|doCV) ? tmpimp.duplicate() : null;
      ImagePlus aveimp = (doAve|doSD|doZS|doCV)       ? tmpimp.duplicate() : null;
      ImagePlus stdimp = (doSD|doZS|doCV)             ? tmpimp.duplicate() : null;
      ImagePlus zsimp  = (doZS)                       ? tmpimp.duplicate() : null;
      ImagePlus cvimp  = (doCV)                       ? tmpimp.duplicate() : null;
      ImagePlus minimp = (doMin)                      ? tmpimp.duplicate() : null;
      ImagePlus maximp = (doMax)                      ? tmpimp.duplicate() : null;
      ImagePlus medimp = (doMed)                      ? tmpimp.duplicate() : null;
      ImagePlus cenimp = (doCen)                      ? tmpimp.duplicate() : null;
      ImagePlus proimp = (doPro|doGeo)                ? tmpimp.duplicate() : null;
      ImagePlus geoimp = (doGeo)                      ? tmpimp.duplicate() : null;
      ImagePlus ssqimp = (doSSq|doMag)                ? tmpimp.duplicate() : null;
      ImagePlus magimp = (doMag)                      ? tmpimp.duplicate() : null;
      ImagePlus cENimp = (doCEN)                      ? tmpimp.duplicate() : null;
      if (doSum|doAve|doSD|doZS|doCV) IJ.run(sumimp, "Set...", "value=0 stack");
      if (doPro|doGeo)                IJ.run(geoimp, "Set...", "value=1 stack");
      if (doSSq|doMag)                IJ.run(ssqimp, "Set...", "value=0 stack");
      if (doSum|doAve|doSD|doZS|doCV|doMin|doMax|doPro|doGeo|doSSq|doMag)
         for(int f=0; f<NF; f++) {
            tmpimp = subhyperstack(subimp,"1-1-1","1-"+NZ+"-1",""+(f+1));
            if (doSum|doAve|doSD|doZS|doCV) ic.run("Add 32-bit stack", sumimp, tmpimp);
            if (doMin)                      ic.run("Min 32-bit stack", minimp, tmpimp);
            if (doMax)                      ic.run("Max 32-bit stack", maximp, tmpimp);
            if (doPro|doGeo) {
               ImagePlus ughimp = tmpimp.duplicate();
               IJ.run(ughimp, "Log", "stack");
               IJ.run(ughimp, "Divide...", "value="+NF+" stack");
               IJ.run(ughimp, "Exp", "stack");
               ic.run("Multiply 32-bit stack", geoimp, ughimp);
               }
            if (doSSq|doMag) {
               ImagePlus ughimp = tmpimp.duplicate();
               IJ.run(ughimp, "Square", "stack");
               ic.run("Add 32-bit stack", ssqimp, ughimp);
               }
            }
      if (doSum) {
         sumimp.setTitle("Sum "+imp.getTitle());
         sumimp.changes = false;
         }
      if (doMin) {
         minimp.setTitle("Minimum "+imp.getTitle());
         minimp.changes = false;
         }
      if (doMax) {
         maximp.setTitle("Maximum "+imp.getTitle());
         maximp.changes = false;
         }
      if (doAve|doSD|doZS|doCV) {
         aveimp = sumimp.duplicate();
         IJ.run(aveimp,"Divide...","value="+NF+" stack");
         aveimp.setTitle("Mean "+imp.getTitle());
         aveimp.changes = false;
         aveimp.setDisplayRange(imp.getDisplayRangeMin(),imp.getDisplayRangeMax());
         }
      if (!doSum) sumimp = null;
      if (doSD|doZS|doCV) {
         stdimp = deviation(subimp,aveimp);
         stdimp.setTitle("Standard Deviation "+imp.getTitle());
         stdimp.changes = false;
         stdimp.resetDisplayRange();
         }
      if (doZS) {
         zsimp = aveimp.duplicate();
         ic.run("Divide 32-bit stack",zsimp,stdimp);
         zsimp.setTitle("Z-Score "+imp.getTitle());
         zsimp.changes = false;
         zsimp.resetDisplayRange();
         }
      if (doCV) {
         cvimp = stdimp.duplicate();
         ic.run("Divide 32-bit stack",cvimp,aveimp);
         cvimp.setTitle("Coefficient of Variation "+imp.getTitle());
         cvimp.changes = false;
         cvimp.resetDisplayRange();
         }
      if (!doSD)  stdimp = null;

      if (doPro) {
         proimp = geoimp.duplicate();
         IJ.run(proimp, "Log", "stack");
         IJ.run(proimp, "Multiply...", "value="+NF+" stack");
         IJ.run(proimp, "Exp", "stack");
         proimp.setTitle("Product "+imp.getTitle());
         proimp.changes = false;
         }
      else proimp = null;
      if (doGeo) {
         geoimp.setTitle("Geometric Mean "+imp.getTitle());
         geoimp.changes = false;
         geoimp.setDisplayRange(imp.getDisplayRangeMin(),imp.getDisplayRangeMax());
         }
      if (doGeo&doDev) {
         ImagePlus devimp = geoimp.duplicate();
         IJ.run(devimp,"Set...","value=1 stack");
         for(int f=0; f<NF; f++) {
            tmpimp = subhyperstack(subimp,"1","1-"+NZ,""+(f+1)).duplicate();
            ic.run("Divide 32-bit stack", tmpimp, geoimp);
            IJ.run(tmpimp, "Log", "stack");
            IJ.run(tmpimp, "Divide...", "value="+NF+" stack");
            IJ.run(tmpimp, "Exp", "stack");
            ic.run("Multiply 32-bit stack", devimp, tmpimp);
            }
         geoimp = (new Concatenator()).concatenate(new ImagePlus[]{geoimp,devimp},true);
         geoimp.setTitle("Geometric Mean and Deviation "+imp.getTitle());
         geoimp.changes = false;
         }
      if (!doGeo)  geoimp = null;

      if (doSSq) {
         ssqimp.setTitle("SumSq "+imp.getTitle());
         ssqimp.changes = false;
         }
      if (doMag) {
         magimp = ssqimp.duplicate();
         IJ.run(magimp,"Square Root","stack");
         magimp.setTitle("Magnitude "+imp.getTitle());
         magimp.changes = false;
         magimp.setDisplayRange(imp.getDisplayRangeMin(),imp.getDisplayRangeMax());
         }
      if (!doSSq) ssqimp = null;

      ImagePlus lrimp  = null;
      ImagePlus tsimp  = null;
      ImagePlus ts2imp = null;
      ImagePlus fitimp = null;
      ImagePlus comimp = null;

      if (doMed|doCen|doLR|doTS|doTS2|doCEN|doFit|doCom) {
         tmpimp = subhyperstack(subimp,"1","1-"+NZ,"1").duplicate();
         IJ.run(tmpimp,"32-bit","");

         final ImagePlus[] lrimps = doLR ? new ImagePlus[4] : new ImagePlus[0];
         for(int i=0; i<lrimps.length; i++)
            lrimps[i] = tmpimp.duplicate();

         final ImagePlus[] tsimps = doTS ? new ImagePlus[2] : new ImagePlus[0];
         for(int i=0; i<tsimps.length; i++)
            tsimps[i] = tmpimp.duplicate();

         final ImagePlus[] ts2imps = doTS2 ? new ImagePlus[3] : new ImagePlus[0];
         for(int i=0; i<ts2imps.length; i++)
            ts2imps[i] = tmpimp.duplicate();

         final ImagePlus[] fpimp = doFit ? new ImagePlus[CurveFitter.getNumParams(fitfun)+2] : new ImagePlus[0];
         for(int i=0; i<fpimp.length; i++)
            fpimp[i] = tmpimp.duplicate();

         final ImagePlus[] caimp = doCom ? new ImagePlus[fp.compute(null,null).length] : new ImagePlus[0];
         for(int i=0; i<caimp.length; i++)
            caimp[i] = tmpimp.duplicate();

         final ImageStack subis = subimp.getStack();

         int[] dims = subimp.getDimensions();
         final int nx = dims[0];
         final int ny = dims[1];
         final int nz = dims[3];
         final int nf = dims[4];

         final double[] X = new double[nf];
         try {
            for(int f=0; f<nf; f++) {
               String str = subis.getSliceLabel(subimp.getStackIndex(1,1,f+1)).split("\n",2)[0].split(",|;",2)[0].split(" = |=| ",2)[1];
               X[f] = Double.valueOf(str).doubleValue();
//               for(int ff=0; ff<f; ff++)
//                  if (X[f] == X[ff])
//                     throw new Throwable();
               }
             }
         catch(Throwable e) {
             for(int f=0; f<nf; f++)
                X[f] = f;
             }


         for(int z=0; z<nz; z++) {
            IJ.showStatus("Computing: z="+z+".."+nz);

            final double[][][] mearr  = new double[nx][ny][1];
            final double[][][] cearr  = new double[nx][ny][1];
            final double[][][] lrarr  = new double[nx][ny][];
            final double[][][] tsarr  = new double[nx][ny][];
            final double[][][] ts2arr = new double[nx][ny][];
            final double[][][] fparr  = new double[nx][ny][];
            final double[][][] caarr  = new double[nx][ny][];
            final double[][][] cEarr  = new double[nx][ny][1];

            final AtomicInteger ai = new AtomicInteger(0);
            final Thread[] threads = new Thread[Runtime.getRuntime().availableProcessors()];
            for(int t=0; t<threads.length; t++) {
               final int Z = z;
               threads[t] = new Thread() {
                  public void run() {
                     double[] Y = new double[nf];
                     for(int x=ai.getAndIncrement(); x<nx; x=ai.getAndIncrement()) {
                        IJ.showProgress(((float)x)/nx);
                        for(int y=0; y<ny; y++) {
                           if ((mask != null) && (mask.getStack().getProcessor(Z+1).get(x,y) != 255))
                              continue;
                           for(int f=0; f<nf; f++) 
                              Y[f] = subis.getProcessor(subimp.getStackIndex(1,Z+1,f+1)).getf(x,y);
                           if (doMed) mearr[x][y][0] = median(Y);
                           if (doCen) cearr[x][y][0] = center(Y);
                           if (doLR)  lrarr[x][y]    = linreg(X,Y);
                           if (doTS)  tsarr[x][y]    = theilsen(X,Y);
                           if (doTS2) ts2arr[x][y]   = theilsen2(X,Y);
                           if (doCEN) cEarr[x][y][0] = centroid(X,Y);
                           if (doFit) {
                              CurveFitter cf = new CurveFitter(X,Y);
                              cf.doFit(fitfun,false);
                              fparr[x][y] = new double[fpimp.length];
                              { int i=0; for(double p : cf.getParams()) fparr[x][y][i++] = p; }
                              fparr[x][y][fpimp.length-2] = cf.getRSquared();
                              fparr[x][y][fpimp.length-1] = cf.getSumResidualsSqr();
                              }
                           if (doCom) caarr[x][y]    = fp.compute(X,Y);
                           }
                        }
                     }
                  };
               threads[t].setPriority(Thread.NORM_PRIORITY);
               threads[t].start();
               }
            try { for(int t=0; t<threads.length; t++) threads[t].join(); }
            catch(InterruptedException ie) { throw new RuntimeException(ie); }

            IJ.showProgress(1F);

            if (doMed) medimp.getStack().setProcessor(arr2ip(mearr,0),z+1);
            if (doCen) cenimp.getStack().setProcessor(arr2ip(cearr,0),z+1);
            if (doLR)  for(int i=0; i<lrimps.length; i++) lrimps[i].getStack().setProcessor(arr2ip(lrarr,i),z+1);
            if (doTS)  for(int i=0; i<tsimps.length; i++) tsimps[i].getStack().setProcessor(arr2ip(tsarr,i),z+1);
            if (doTS2) for(int i=0; i<ts2imps.length; i++) ts2imps[i].getStack().setProcessor(arr2ip(ts2arr,i),z+1);
            if (doFit) for(int i=0; i<fpimp.length;  i++) fpimp[i].getStack().setProcessor(arr2ip(fparr,i),z+1);
            if (doCom) for(int i=0; i<caimp.length;  i++) caimp[i].getStack().setProcessor(arr2ip(caarr,i),z+1);
            if (doCEN) cENimp.getStack().setProcessor(arr2ip(cEarr,0),z+1);
            }
         if (doMed) {
            medimp.setTitle("Median "+imp.getTitle());
            medimp.setDisplayRange(imp.getDisplayRangeMin(),imp.getDisplayRangeMax());
            medimp.changes = false;
            }
         if (doCen) {
            cenimp.setTitle("Center "+imp.getTitle());
            cenimp.setDisplayRange(imp.getDisplayRangeMin(),imp.getDisplayRangeMax());
            cenimp.changes = false;
            }
         if (doLR) {
            lrimp = (new Concatenator()).concatenate(lrimps,true);
            lrimp.setTitle("Liner Regression "+imp.getTitle());
            lrimp.changes = false;
            }
         if (doTS) {
            tsimp = (new Concatenator()).concatenate(tsimps,true);
            tsimp.setTitle("Theil Sen Regression "+imp.getTitle());
            tsimp.changes = false;
            }
         if (doTS2) {
            ts2imp = (new Concatenator()).concatenate(ts2imps,true);
            ts2imp.setTitle("Theil Sen 2 Regression "+imp.getTitle());
            ts2imp.changes = false;
            }
         if (doCen) {
            cENimp.setTitle("Centoid "+imp.getTitle());
            cENimp.setDisplayRange(imp.getDisplayRangeMin(),imp.getDisplayRangeMax());
            cENimp.changes = false;
            }
         if (doFit) {
            fitimp = (new Concatenator()).concatenate(fpimp,true);
            fitimp.setTitle(CurveFitter.fitList[CurveFitter.sortedTypes[fitfun]]+" Regression "+imp.getTitle());
            fitimp.changes = false;
            }
         if (doCom) {
            comimp = (new Concatenator()).concatenate(caimp,true);
            comimp.setTitle("Compute "+imp.getTitle());
            comimp.changes = false;
            }
         }

      if (doAve&doDev) {
         aveimp.setOpenAsHyperStack(true);
         ImagePlus devimp = deviation(subimp,aveimp);
         aveimp = (new Concatenator()).concatenate(new ImagePlus[]{aveimp,devimp},true);
         aveimp.setTitle("Mean and Deviation "+imp.getTitle());
         aveimp.changes = false;
         }
      if (doMin&doDev) {
         minimp.setOpenAsHyperStack(true);
         ImagePlus devimp = deviation(subimp,minimp);
         minimp = (new Concatenator()).concatenate(new ImagePlus[]{minimp,devimp},true);
         minimp.setTitle("Minimum and Deviation "+imp.getTitle());
         minimp.changes = false;
         }
      if (doMax&doDev) {
         maximp.setOpenAsHyperStack(true);
         ImagePlus devimp = deviation(subimp,maximp);
         maximp = (new Concatenator()).concatenate(new ImagePlus[]{maximp,devimp},true);
         maximp.setTitle("Maximum and Deviation "+imp.getTitle());
         maximp.changes = false;
         }
      if (doMed&doDev) {
         medimp.setOpenAsHyperStack(true);
         ImagePlus devimp = deviation(subimp,medimp);
         medimp = (new Concatenator()).concatenate(new ImagePlus[]{medimp,devimp},true);
         medimp.setTitle("Mediam and Deviation "+imp.getTitle());
         medimp.changes = false;
         } 

      if (doCen&doDev) {
         cenimp.setOpenAsHyperStack(true);
         ImagePlus devimp = deviation(subimp,cenimp);
         cenimp = (new Concatenator()).concatenate(new ImagePlus[]{cenimp,devimp},true);
         cenimp.setTitle("Center and Deviation "+imp.getTitle());
         cenimp.changes = false;
         }

      if (!doAve) aveimp = null;

      return new ImagePlus[]{ sumimp, aveimp, stdimp, zsimp, cvimp, minimp, maximp, medimp, cenimp, proimp, geoimp, ssqimp, magimp, lrimp, tsimp, ts2imp, cENimp, fitimp, comimp};
      }

   private static FloatProcessor arr2ip(double[][][] full, int i) {
      float[][] arr = new float[full.length][full[0].length];
      for(int x=0; x<full.length; x++)
         for(int y=0; y<full[0].length; y++)
            arr[x][y] = full[x][y] == null ? Float.NaN : (float)full[x][y][i];
      return new FloatProcessor(arr);
      }

   private static ImagePlus deviation(ImagePlus theimp, ImagePlus refimp) {
      ImageCalculator ic = new ImageCalculator();
      int NF = theimp.getNFrames();
      int NZ = theimp.getNSlices();
      ImagePlus devimp = refimp.duplicate();
      IJ.run(devimp,"Set...","value=0 stack");
      for(int f=0; f<NF; f++) {
         ImagePlus tmpimp = subhyperstack(theimp,"1","1-"+NZ,""+(f+1)).duplicate();
         ic.run("Subtract 32-bit stack", tmpimp, refimp);
         IJ.run(tmpimp,"Square","stack");
         ic.run("Add 32-bit stack", devimp, tmpimp);
         }
      IJ.run(devimp,"Square Root","stack");
      IJ.run(devimp,"Divide...","value="+NF+" stack");
      devimp.setOpenAsHyperStack(true);
      return devimp;
      }

   /** append frames onto a hyperstack.
    *  use at your own risk.
    */
   public static void append(ImagePlus ... imps) {
      if (imps.length < 2) return;
      ImageStack is0 = imps[0].getStack();
      for(int i=1; i<imps.length; i++) {
         ImageStack isi = imps[i].getStack();
         int ns = imps[i].getImageStackSize();
         for(int s=0; s<ns; s++)
            is0.addSlice(isi.getSliceLabel(s+1),isi.getProcessor(s+1));
         }
      }
   private static ImagePlus concat(ImagePlus theimp, ImagePlus appimp) {
      ImagePlus imp = theimp.duplicate();
      ImageStack is1 = imp.getStack();
      ImageStack is2 = appimp.getStack();
      for(int s=0; s<appimp.getImageStackSize(); s++)
         is1.addSlice(is2.getSliceLabel(s+1),is2.getProcessor(s+1));
      return imp;
      }

   /** and copy the metadata also.
    */
   public static ImagePlus subhyperstack(ImagePlus imp, String cl, String zl, String fl) {
      ImagePlus simp = SubHyperstackMaker.makeSubhyperstack(imp,cl,zl,fl);
      simp.setTitle(imp.getTitle());
      simp.setCalibration(imp.getCalibration());
      simp.setFileInfo(imp.getOriginalFileInfo());
      simp.setProperty("Info",imp.getProperty("Info"));
      simp.setDisplayRange(imp.getDisplayRangeMin(),imp.getDisplayRangeMax());
      return simp;
      }

   /** weighted mean weighted to the median
    */
   public static double center(double[] a) {
      if (a == null || a.length == 0) return Double.NaN;
      if (a.length == 1) return a[0];
      double M = median(a);
      if (a.length < 5) return M;

      double[] D = new double[a.length];
      for(int i=0; i<a.length; i++)
         D[i] = (a[i] > M) ? a[i]-M : M-a[i];
      double d = median(D);
      
      double wms = 0.0;
      double ws  = 0.0;
      for(int i=0; i<a.length; i++) {
         double amMamM = (a[i]-M)*(a[i]-M);
         if (amMamM > 5*d) continue;
         double w = 1.0/(1.0+amMamM);
         wms += w*a[i];
         ws  += w;
         }
      return(wms/ws);
      }

//   public static double[] theilsen2(double[] x, double[] y) {
   /** theilsen regressor for quadradic.
    *  can be called as x[],y[] or xy[][] but only the first pair computed
    */
   public static double[] theilsen2(double[] ... _x) {
      if (_x.length < 2) return null;
      double[] x = _x[0];
      double[] y = _x[1];

      double a=Double.NaN, b=Double.NaN, c=Double.NaN;
      {
         Vector<Double> va = new Vector<Double>(2*x.length, x.length);
         for(int i=0; i<x.length; i++)
            for(int j=i+1; j<x.length; j++)
               for(int k=j+1; k<x.length; k++)
                  va.add( (x[k]*(y[j]-y[i])+x[j]*(y[i]-y[k])+x[i]*(y[k]-y[j]))
                         / ((x[i]-x[j])*(x[i]-x[k])*(x[j]-x[k])) );
         a = center(va.toArray(new Double[0]));
      }

      double[] ymaxx = x.clone();
      for(int i=0; i<x.length; i++)
         ymaxx[i] = y[i] - a*x[i]*x[i];

      {
         Vector<Double> vb = new Vector<Double>(2*x.length, x.length);
         for(int i=0; i<x.length; i++)
            for(int j=i+1; j<x.length; j++)
               vb.add( (ymaxx[j]-ymaxx[i]) / (x[j]-x[i]) );
         b = center(vb.toArray(new Double[0]));
      }

      {
         Double[] vc = new Double[x.length];
         for(int i=0; i<x.length; i++)
            vc[i] = ymaxx[i] - b*x[i];
         c = center(vc);
      }

      return new double[]{a,b,c};
      }

   private static double median(Double[] a) {
      if (a == null || a.length == 0) return Double.NaN;
      if (a.length == 1) return a[0];
      double[] tmp = new double[a.length];
      for(int i=0; i<a.length; i++)
         tmp[i] = a[i];
      Arrays.sort(tmp);
      return (tmp.length%2 == 1) ? tmp[tmp.length/2] : (tmp[tmp.length/2-1]+tmp[tmp.length/2])/2.0;
      }

   public static double center(Double[] a) {
      if (a == null || a.length == 0) return Double.NaN;
      if (a.length == 1) return a[0];
      double M = median(a);
      if (a.length < 5) return M;

      double[] D = new double[a.length];
      for(int i=0; i<a.length; i++)
         D[i] = (a[i] > M) ? a[i]-M : M-a[i];
      double d = median(D);
      
      double wms = 0.0;
      double ws  = 0.0;
      for(int i=0; i<a.length; i++) {
         double amMamM = (a[i]-M)*(a[i]-M);
         if (amMamM > 5*d) continue;
         double w = 1.0/(1.0+amMamM);
         wms += w*a[i];
         ws  += w;
         }
      return(wms/ws);
      }

//   public static double[] theilsen(double[] x, double[] y) {
   /** theilsen regressor for linear.
    *  can be called as x[],y[] or xy[][] but only the first pair computed
    */
   public static double[] theilsen(double[] ... _x) {
      if (_x.length < 2) return null;
      double[] x = _x[0];
      double[] y = _x[1];

      int nf = x.length;

      int np = 0;
      for(int i=0; i<(nf-1); i++)
         np += (nf-i-1);

      double[] yp = new double[np];
      for(int i=0,k=0; i<(nf-1); i++)
         for(int j=i+1; j<nf; j++,k++)
            yp[k] =((y[i]-y[j])/(x[i]-x[j]));
      double M = center(yp);

      double[] xp = new double[np];
      for(int i=0,k=0; i<(nf-1); i++)
         for(int j=i+1; j<nf; j++,k++)
            xp[k] = y[i] - M*x[i];
      double B = center(xp);

      return(new double[]{M,B});
      }

//   public static double[] linreg(double[] x, double[] y) {
   /** linear least squares regressor
    *  can be called as x[],y[] or xy[][] but only the first pair computed
    */
   public static double[] linreg(double[] ... _x) {
      if (_x.length < 2) return null;
      double[] x = _x[0];
      double[] y = _x[1];

      double xbar = 0; for(double v : x) xbar += v; xbar /= x.length;
      double ybar = 0; for(double v : y) ybar += v; ybar /= y.length;

      double xxbar = 0, yybar = 0, xybar = 0;
      for(int i=0; i<x.length; i++) {
         double dx = (x[i]-xbar), dy = (y[i]-ybar);
         xxbar += dx*dx;
         yybar += dy*dy;
         xybar += dx*dy;
         }

      double m = xybar/xxbar;
      double b = ybar-m*xbar;
      
      double ssr = 0;
      for(int i=0; i<x.length; i++) {
         double Yi = m*x[i]+b;
         ssr += (y[i] - Yi)*(y[i] - Yi);
         }
      double rs = 1.0 - (ssr/yybar);

      double rxy = xybar/Math.sqrt(xxbar*yybar);

      return new double[]{m,b,rs,rxy};
      }

   private static double median(double[] arr) {
      double[] tmp = new double[arr.length];
      System.arraycopy(arr, 0, tmp, 0, arr.length);
      Arrays.sort(tmp);
      return (tmp.length%2 == 1) ? tmp[tmp.length/2] : (tmp[tmp.length/2-1]+tmp[tmp.length/2])/2;
      }

   /** centroid 
    *  can be called as x[],y[]...
    */
   public static double[] centroid(double[] ... x) {
      if (x.length < 2) return null;
      double[] res = new double[x.length-1];
      for(int i=0; i<x.length-1; i++)
         res[i] = centroid(x[i],x[x.length-1-1]);
      return res;
      }
   public static double centroid(double[] x, double[] y) {
      double sumxy = 0, sumy = 0;
      for(int i=0; i<x.length; i++) {
         sumxy += x[i]*y[i];
         sumy  += y[i];
         }
      return sumxy/sumy;
      }

   private static float[] d2f(double[] d) {
      if (d==null) return null;
      float[] f = new float[d.length];
      for(int i=0; i<d.length; i++)
         f[i] = (float)d[i];
      return f;
      }
   private static double[] f2d(float[] f) {
      if (f==null) return null;
      double[] d = new double[f.length];
      for(int i=0; i<d.length; i++)
         d[i] = (double)f[i];
      return d;
      }
   private static double[] D2d(Double[] D) {
      double[] d = new double[D.length];
      for(int i=0; i<d.length; i++)
         d[i] = D[i]; 
      return d;
      }
}
